# * ====================================================================
# * Copyright (c) 1995,1996, ISIS/JRC.  All rights reserved.
# *
# * This product was developed within the Institute for Safety, Informatics
# * and Systems of the Joint Research Centre of the European Commission.
# *
# * More information about us; and this software product can
# * be found on http://www.jrc.it.
# *
# * With this file you should have received a file called 'License'.
# *
# * Redistribution and use in source and binary forms, with or without
# * modification, are permitted provided that the following conditions
# * are met:
# *
# * 1. Redistributions of source code must retain the above copyright
# *    notice, this list of conditions and the following disclaimer.
# *
# * 2. Redistributions in binary form must reproduce the above copyright
# *    notice, this list of conditions and the following disclaimer in
# *    the documentation and/or other materials provided with the
# *    distribution.
# *
# * 3. All materials mentioning features or use of this
# *    software must display the following acknowledgment:
# *    "This product includes software developed by ISIS/JRC"
# *
# * 4. The name of the Copyright holder must not be used to
# *    endorse or promote products derived from this software without
# *    prior written permission.
# *
# * 5. Redistributions of any form whatsoever must retain the following
# *    acknowledgment:
# *    "This product includes software developed by ISIS/JRC"
# *
# * THIS SOFTWARE IS PROVIDED BY ISIS/JRC ``AS IS'' AND ANY
# * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL ISIS/JRC OR
# * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# * OF THE POSSIBILITY OF SUCH DAMAGE.
# * ====================================================================
#
# Type 'perldoc Z3950' to see the documentation.
#
# History:
#  1999-02-26  NWH   major reshuffling of code into different
#                    modules and files; Initial release

package Z3950;

use strict;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK);

require Exporter;

@ISA = qw(Exporter);
@EXPORT = qw();
$VERSION = '0.3';

=head1 NAME

Z3950 - Perl extension to create Z3950 clients

=head1 SYNOPSIS

   use Z3950;

   $Z3950::CONNECTION_LIFETIME = 1800;
   $Z3950::DEFAULT_PORT = 210;

   ($handle, %PDU) = &Z3950::InitConnection(
         host=>'ceo.gelos.org', port=2210, type=>'BER');
   die 'ceo.gelos.org:2210:BER '.$@
      if !$handle;

   &Z3950::TimeStampConnection($handle);
   &Z3950::CheckConnection($handle);
   &Z3950::CheckConnections()
      or die "No connections open anymore;
   &Z3950::CloseConnection($handle);
   &Z3950::CloseConnections();


   $attr = &Z3950::Attribute($type, $value);
   $attr = &Z3950::Field($field, {anywhere=>1035});
   $attr = &Z3950::Relation($field, {relevance=>102});

   $operand = &Z3950::RPNOperand('and', @attributes);
   $operand = &Z3950::RPNAnd($operand1, $operand2);
   $operand = &Z3950::RPNOr($operand1, $operand2);
   $operand = &Z3950::RPNStructure(2, $operand1, $operand2);
   $query = &Z3950::RPNCreate($attrset, $operand);


   %SR = &Z3950::Search($handle, 'dbase name', $query,
            $fullnresults, $briefnresults, maxnresults);
      or warn 'ceo.gelos.org:2210: '.$@;

   %PR = &Z3950::Present($handle, $full, $firstresult, $nresults)
      or warn 'ceo.gelos.org:2210: '.$@;

   $errmsg = &Z3950::diagnostic($diagnostic, 'default message');

   &Z3950::parseRecord($record, $quirks, 'record');

=head1 DESCRIPTION

The Z39.50 enables you to query remote z39.50 databases.

=head1 AUTHOR

Nick Hibma <nick.hibma@jrc.it>

=head2 Variables & Functions

=over 3

=cut

use strict;

use Socket;
use FileHandle;

use BER;
use ASN1;
use Z3950::Z3950;
use Z3950::ASN1::Z39_50_APDU_1995;

# Initialisation of Z3950-global variables.
#

# %Connections
#  hash, containing (apart from the items described in
#  the header of InitConnection):
#     handle      Socket handle
#     errmsg      Contains error message in case of error
#     timestamp   time of last use of the connection
# XXX Is it possible to get rid of this hash?
#     Might make it possible to drop the connection
# TODO Turn the whole connection thingie into objects ?
#
%Z3950::Connections = ();


=item $CONNECTION_LIFETIME

Lifetime of a connection to a Z server. After this amount of
time of no activity the connection is closed.

=cut

$Z3950::CONNECTION_LIFETIME = 1800;

=item $DEFAULT_PORT

Default port for Z39.50 connections (210)

=cut

$Z3950::DEFAULT_PORT = 210;


# Z3950 functionality

=item InitConnection($serverconfig)


Initializes a connection to a remote host.

   $serverconfig     reference to a hash with the following elements
      host           hostname or IP address (required)
      port           port number or the string 'z39.50'
      type           one of 'XER', 'BER' (supported ASN.1 encodings)
      quirks         quirks for the server
         nonamedresultsets    does not support Named Result Sets
         nofullelemset        does not return full elements in SearchResponse
      recordsyntax   list of OID's of record syntaxes supported
      dbase          databases used on this server
                     reference to a hash, ->{dbase}->{<dbasename>}:
         quirks      quirks for the databases
            dashtospace       convert [-_] to a space
            correctcase       convert the first character of a word to uppercase
            htmlcontent       the content of a value is html
            keyvaluetree      the SUTR record is formatted like a keyvaluetree
            coloncentered     the SUTR record is formatted centered on the colon
            dif               the SUTR record is formatted as a DIF record

Apart from carefull modifications to the dbase quirks, nothing should be
changed in the hash after initialisation of the connection!

Returns ($handle, $InitializeResponse), or, if !wantarray, $handle, or,
in case of error, undef. In that case $@ contains the error string.

=cut

sub InitConnection {
   my ($serverconfig) = @_;

   # Check the hash for the necessary bits and pieces and
   # add the defaults if they are missing
   #
   if ( !$serverconfig->{host} ) {
      $@ = 'InitConnection: hostname not specified';
      return undef;
   }

   $serverconfig->{port} = $Z3950::DEFAULT_PORT
      if !$serverconfig->{port} or $serverconfig->{port} =~ m/z39\.?50/;
   $serverconfig->{type} = 'BER'
      if !$serverconfig->{type};

   my $handle = $serverconfig->{handle} = new FileHandle;

   # Create the info on the connection
   # The timestamp is used to later zap the connection if it has been unused
   # for a while.

   $Z3950::Connections{$handle} = {
               %{ $serverconfig },
               timestamp => time(),
            };
   warn "-- Initialising connection to tcp:$serverconfig->{host}:$serverconfig->{port}.$serverconfig->{type}"
      if $Z3950::debug;

   if ( !socket($handle, PF_INET, SOCK_STREAM, getprotobyname('tcp')) ) {
      $@ = 'socket failed, ' . $!;
      &Z3950::CloseConnection($handle);
      return undef;
   }

   if ( !connect($handle, sockaddr_in($serverconfig->{port},inet_aton($serverconfig->{host}))) ) {
      $@ = 'connect failed, ' . $!;
      &Z3950::CloseConnection($handle);
      return undef;
   }

   # Build up the association with the server
   #
   # The specified preferredMessageSize and exceptionalRecordSize are chosen
   # as big as the response given by Isite (arbitrary choice that is) and
   # could be chosen bigger. Trade-off between use of memory and reducing
   # the number of transfers required to transfer results.
   #

   my $PDU = undef;
   $PDU->{initRequest}->{protocolVersion} = "version-1 version-2";
   $PDU->{initRequest}->{options} = "search present namedResultSets";
   $PDU->{initRequest}->{preferredMessageSize} = 1048756;
   $PDU->{initRequest}->{exceptionalRecordSize} = 10487560;
   $PDU->{initRequest}->{implementationId} = $::IDSTRING;
   $PDU->{initRequest}->{implementationName} = $::NAMESTRING;
   $PDU->{initRequest}->{implementationVersion} = $::VERSIONSTRING;

   if ( !&ASN1::sendPDU($handle, $PDU, \%Z3950::ASN1::Z39_50_APDU_1995::PDU,
                       $serverconfig->{type}, 'InitializeRequest') ) {
      $@ = 'Send of initRequest failed (' . $@ . ')';
      &Z3950::CloseConnection($handle);
      return undef;
   }


   # Read a response from the server
   #
   $PDU = &ASN1::receivePDU($handle, \%Z3950::ASN1::Z39_50_APDU_1995::PDU,
                            $serverconfig->{type}, 'InitializeResponse');
   if ( !$PDU ) {
      $@ = 'Receive of initResponse failed (' . $@ . ')';
      &Z3950::CloseConnection($handle);
      return undef;
   }
   if ( !exists $PDU->{initResponse} ) {
      $@ = 'Wrong tag ('.join(', ', keys %{ $PDU }).'), not initResponse';
      &Z3950::CloseConnection($handle);
      return undef;
   }

   if ( $PDU->{initResponse}->{options} !~ m/\bsearch\b/
        or $PDU->{initResponse}->{options} !~ m/\bpresent\b/ ) {
       $@ = 'Search or Present not supported (IS options: \''
          . $PDU->{initResponse}->{options}
          . '\')?! What silly server is this?';
       &Z3950::CloseConnection($handle);
       return undef;
   }

   $Z3950::Connections{$handle}->{quirks}->{nonamedresultsets} = 1
      if $PDU->{initResponse}->{options} !~ m/\bnamedResultSets\n/;


   if ( $Z3950::debug ) {
      warn '-- Name: ' . $PDU->{initResponse}->{implementationName}
         . ' (Id = '. $PDU->{initResponse}->{implementationId}
         . '), version '.$PDU->{initResponse}->{implementationVersion};
      warn '-- Supports options: ' . $PDU->{initResponse}->{options};
   }

   return (wantarray? ($handle, $PDU->{initResponse}) : $handle);
}


=item Search($handle, $dbase, $query, $fullnresults, $briefnresults)

Perform a search on a remote server.

   $handle           connection handle
   $dbase            name of the database to query
   $query            Query, as returned by RPNCreate
   $fullnresults     maximum number of full records to return [1..]
   $briefnresults    maximum number of brief records to return [1..]
   $maxnresults      maximum size of result set [1..]

Returns SearchResponse hash, or, in case of error, undef ($@ contains error string).

=cut

sub Search {
   my ($handle, $dbase, $query, $fullnresults, $briefnresults, $maxnresults) = @_;

   $maxnresults = 1000
      if !defined($maxnresults);
   $maxnresults++;                        # largeSetLowerBound is a lower limit not an upper.

   # Prepare the parameters for the SearchRequest
   #
   $maxnresults = $fullnresults++         # largeSetLowerBound >= smallSetUpperBound+1
      if $maxnresults <= $fullnresults;
   $maxnresults = $briefnresults
      if $maxnresults < $briefnresults;

   # FIXME ResultSetId (ResultSetName) is hardcoded to 'default'
   my $resultset = 'default';

   # default to SUTRS of no preferred record syntax is available
   #
   my $recordsyntax = $Z3950::OID_RS_SUTRS;
   $recordsyntax = $Z3950::Connections{$handle}->{recordsyntax}->[0]
      if $Z3950::Connections{$handle}->{recordsyntax}->[0];

   $fullnresults = 0
      if $Z3950::Connections{$handle}->{quirks}->{nofullelemset};

   warn '-- Performing search'
      . ($briefnresults? ", present brief 1+$briefnresults":'')
      . ($fullnresults? ", present full 1+$fullnresults":'')
      . ": Resultset '$resultset', "
      . "Sets: $fullnresults/$briefnresults/$maxnresults, "
      . 'RSyntax: \'' . &ASN1::OIDdescription($recordsyntax) . '\''
      if $Z3950::debug;

   # Build the search request and send it
   #
   my $PDU = undef;
   $PDU->{searchRequest}->{smallSetUpperBound} = $fullnresults;
   $PDU->{searchRequest}->{largeSetLowerBound} = $maxnresults;
   $PDU->{searchRequest}->{mediumSetPresentNumber} = $briefnresults;
   $PDU->{searchRequest}->{replaceIndicator} = $BER::TRUE;
   $PDU->{searchRequest}->{resultSetName} = $resultset;
   $PDU->{searchRequest}->{databaseNames}->[0] = $dbase;
   # FIXME element set names are hardcoded to brief, full respectively
   $PDU->{searchRequest}->{smallSetElementSetNames}->{genericElementSetName} = 'f';
   $PDU->{searchRequest}->{mediumSetElementSetNames}->{genericElementSetName} = 'b';
   $PDU->{searchRequest}->{preferredRecordSyntax} = $recordsyntax;
   $PDU->{searchRequest}->{query} = $query;

   &ASN1::sendPDU($handle, $PDU, \%Z3950::ASN1::Z39_50_APDU_1995::PDU,
                  $Z3950::Connections{$handle}->{type}, 'SearchRequest')
      or return undef;

   # Wait for return packet
   #
   $PDU = &ASN1::receivePDU($handle, \%Z3950::ASN1::Z39_50_APDU_1995::PDU,
                            $Z3950::Connections{$handle}->{type}, 'SearchResponse')
      or return undef;

   if ( !exists $PDU->{searchResponse} ) {
      $@ = 'Wrong tag ('.join(', ', keys %{ $PDU }).'), not searchResponse';
      return undef;
   }

   return $PDU->{searchResponse};
}


=item Present($handle, $full, $firstresult, $nresults)

Perform a present on a remote server for results found with Search().

   $handle           connection handle
   $full             should we retrieve full or brief records
   $firstresult      index of first result to retrieve ([1..])
   $nresults         number of results to retrieve

Returns PresentResponse hash, or, in case of error, undef ($@ contains error string).

=cut

sub Present {
   my ($handle, $full, $firstresult, $nresults) = @_;

   # Prepare parameters for PresentRequest
   #
   my $full = ($full? 'F' : 'B');

   warn '$firstresult = 0, but should be in range [1..], base 0 instead of 1 problem?'
      if $Z3950::debug and $firstresult == 0;
   $firstresult = 1
      if $firstresult == 0;

   # FIXME ResultSetId (ResultSetName) is hardcoded to 'default'
   my $resultset = 'default';

   my $recordsyntax = $Z3950::OID_RS_SUTRS;
   $recordsyntax = $Z3950::Connectons{$handle}->{recordsyntax}->[0]
      if $Z3950::Connectons{$handle}->{recordsyntax}->[0];

   warn "-- Performing Present $firstresult+$nresults, "
      . "ResultSet: $resultset, RSyntax=" . &ASN1::OIDdescription($recordsyntax)
      if $Z3950::debug;

   # Build the present request and send it
   #
   #
   my $PDU = undef;
   $PDU->{presentRequest}->{resultSetId} = $resultset;
   $PDU->{presentRequest}->{resultSetStartPoint} = $firstresult;
   $PDU->{presentRequest}->{numberOfRecordsRequested} = $nresults;
   $PDU->{presentRequest}->{recordComposition}->{simple}->{genericElementSetName} = $full;
   $PDU->{presentRequest}->{preferredRecordSyntax} = $recordsyntax;

   # Send off present request to Z39.50 server
   #
   &ASN1::sendPDU($handle, $PDU, \%Z3950::ASN1::Z39_50_APDU_1995::PDU, $Z3950::Connections{$handle}->{type}, 'PresentRequest')
      or return undef;

   # Wait for return packet
   #
   my $PDU = &ASN1::receivePDU($handle, \%Z3950::ASN1::Z39_50_APDU_1995::PDU, $Z3950::Connections{$handle}->{type}, 'PresentResponse')
      or return undef;
   if ( !exists $PDU->{presentResponse} ) {
      $@ = 'Wrong tag ('.join(', ', keys %{ $PDU }).'), not presentResponse';
      return undef;
   }

   return $PDU->{presentResponse};
}


=item CloseConnection(@handles)

Shutdown the connection to the Z39.50-server.

=item CloseConnections()

Closes all connections presently open.

=cut

sub CloseConnection {
   my (@handles) = @_;
   my $handle;

   foreach $handle (@handles) {
      if ( exists $Z3950::Connections{$handle} ) {
         warn '-- Closing connection to '
            . $Z3950::Connections{$handle}->{host}
            . ':' . $Z3950::Connections{$handle}->{port}
            . '.' . $Z3950::Connections{$handle}->{type}
            if $Z3950::debug;
         # A bit over the top, but let's make sure we kill the filehandle
         delete $Z3950::Connections{$handle}->{handle};
         delete $Z3950::Connections{$handle};
      }
   }
}

sub CloseConnections {
   &CloseConnection(keys %Z3950::Connections);
}


=item TimeStampConnection(@handles)

Renew timestamp on connection.

=item CheckConnection(@handles)

Check the age of the connection, and close if it is
older than $Z3950::CONNECTION_LIFETIME.

=item CheckConnections()

Check the age of all connections, and close if it is
older than $Z3950::CONNECTION_LIFETIME.

Returns the number of still open connections.

=cut

sub TimeStampConnection {
   my (@handles) = @_;
   my $handle;

   foreach $handle (@handles) {
      $Z3950::Connections{$handle}->{timestamp} = time();
   }
}

sub CheckConnection {
   my (@handles) = @_;
   my $handle;

   # We could consider sending a dummy packet to the other side to make sure
   # the other side is still alive.
   #
   foreach $handle (@handles) {
      next
         if !$handle;
      warn '-- Checking connection to '
         . $Z3950::Connections{$handle}->{host}.':'.$Z3950::Connections{$handle}->{port}
         if $Z3950::debug;
      &CloseConnection($handle)
         if time()-$Z3950::Connections{$handle}->{timestamp} > $Z3950::CONNECTION_LIFETIME;
   }
}

sub CheckConnections {
   &CheckConnection(keys %Z3950::Connections);

   warn '-- '.(scalar(keys %Z3950::Connections)).' connections still open'
      if $Z3950::debug;
   return scalar(keys %Z3950::Connections);
}




=head2 TYPE-1 QUERY

The functions below are used to create Type-1 queries.
The functions and their return values should be matched according
to the types they require as input and return.

=item RPNCreate($attrset, $operand)

Creates an RPN query (type-1).

   $attrset       OID of attribute set being used
   $operand       RPNStructure

Returns Query.

=item RPNAnd(@operands)

Creates and relation between all the operands.

   @operands      RPNStructure

Returns RPNStructure.

=item RPNOr(@operands)

Creates or relation between all the operands.

   @operands      RPNstructure

Returns RPNStructure.

=item RPNStructure($operator, @operands)

Creates a relation between the operands.

   $operator      one of 'and', 'or', 'and-not', 'prox'
   @operands      RPNstructure

Returns RPNstructure.

=item RPNOperand($term, @attributes)

Creates an operand.

   $term          search term (CHOICE general in Term)
   @attributes    list of AttributeElement

Returns RPNStructure.

=cut

sub RPNCreate {
   my ($attributeset, $operand) = @_;

   return undef
      if !$operand;

   return {'type-1'=>{ 'attributeSet'=>$attributeset, 'rpn'=>$operand}};
}

sub RPNAnd {
   my (@operands) = @_;

   return &RPNStructure('and', @operands);
}

sub RPNOr {
   my (@operands) = @_;

   return &RPNStructure('or', @operands);
}

sub RPNStructure {
   my ($operator, @operands) = @_;     # operation tag and the sequence of operands
   my ($r, $rt);

   @operands = grep { $_ } @operands;     # Filter out empty ones

   return undef                     # nothing there
      if $#operands == -1;
   return $operands[0]              # only one, then leave it as it is
      if $#operands == 0;

   $r = shift @operands;

   while ( $rt = shift @operands ) {
      $r = {'rpnRpnOp'=>{
               'rpn1'=>$r,
               'rpn2'=>$rt,
               'op'=>{
                  $operator=>'defined'}}};
   }

   return $r;
}

sub RPNOperand {
   my ($term, @attributes) = @_;

   @attributes = grep { $_ } @attributes;    # filter out empty ones

   return undef
      if $#attributes == -1;

   return {'op'=>{
             'attrTerm'=>{
                'attributes'=>[@attributes],
                'term'=>{
                   'general'=>$term}}}};
}


=item Attribute($type, $value)

Translate from a name or number to an attribute.

   $type    attribute type, numeric id
   $value   attribute value, numeric id

Returns AttributeElement.

=item Field($field, @attrlists)

Translate from a name or number to a field attribute
(BIB/1 Use attribute, AttributeType=1).

   $field      name of field or numeric id
   @attrlists  list of {'anywhere'=>1035}

Returns AttributeElement.

=item Relation($field, @attrlists)

Translate from a name or number to a relation attribute
(BIB/1 Relation attribute, AttributeType=2).

   $field      name of the field or numeric id
   @attrlists  list of {'relevance'=>102}

Returns AttributeElement.

=cut

sub Attribute {
   my ($type, $value) = @_;

   return undef
      unless defined($type) and defined($value);

   return {'attributeType'=>$type,
           'attributeValue'=>{
              'numeric'=>$value}};
}

sub Field {
   my ($field, @attrlists) = @_;

   return undef
      unless defined $field;

   my $fieldname = $field
      if $Z3950::debug;

   my $found = 0;

   if ( $field =~ m/^\d+$/ ) {
      # Don't do anything with a numeric ID
      #
      $found = 1;
   } else {
      $field = lc($field);
      # Try to look the field up in the list of fieldlists provided
      # in the description of the server
      #
      my $fieldlist;
      foreach $fieldlist ( (@attrlists, \%Z3950::Bib1_Use) ) {
         if ( defined($fieldlist->{$field}) ) {
            $field = $fieldlist->{$field};
            $found = 1;
            last;
         }
      }
   }

   if ( $found ) {
      warn "-- Field $fieldname translated to $field"
         if $Z3950::debug;
      return &Attribute(1, $field);
   } else {
      warn "-- Field $fieldname not found"
         if $Z3950::debug;
      return undef;
   }
}

sub Relation {
   my ($relation, @attrlists) = @_;

   return undef
      unless defined $relation;

   my $relationname = $relation
      if $Z3950::debug;

   my $found = 0;

   if ( $relation =~ m/^\d+$/ ) {
      # Don't do anything with a numeric Id
      $found = 1;
   } else {
      # Try to look the relation up in the list of relationlists provided
      # in the description of the server, and the Bib1_Use list.
      #
      my $relationlist;
      foreach $relationlist ( (@attrlists, $Z3950::Bib1_Relation) ) {
         if ( defined($relationlist->{$relation}) ) {
            $relation = $relationlist->{$relation};
            $found = 1;
            last;
         }
      }
   }

   if ( $relation ) {
      warn "-- Relation $relationname translated to $relation"
         if $Z3950::debug;
      return &Attribute(2, $relation);
   } else {
      warn "-- relation $relationname not found"
         if $Z3950::debug;
      return undef;
   }
}


# The parsing of returned records in an EXTERNAL field
#


# parseIntUnit($intUnit)
#     Parses an IntUnit structure as specified in Z39.50-APDU-1995
#     Returns its value.
#
sub parseIntUnit {
   my ($intUnit) = @_;

   my $value = $intUnit->{value};

   warn '-- Ignoring $intUnit->{unitUsed}->{unitSystem} = \''.$intUnit->{unitUsed}->{unitSystem}.'\''
      if $Z3950::debug and defined($intUnit->{unitUsed}->{unitSystem});
   warn '-- Ignoring $intUnit->{unitUsed}->{unitType}->{string} = \''
      . $intUnit->{unitUsed}->{unitType}->{string}.'\''
      if $Z3950::debug and defined($intUnit->{unitUsed}->{unitType}->{string});
   warn '-- Ignoring $intUnit->{unitUsed}->{unitType}->{numeric} = \''
      . $intUnit->{unitUsed}->{unitType}->{numeric}.'\''
      if $Z3950::debug and defined($intUnit->{unitUsed}->{unitType}->{numeric});

   $value *= 10**$intUnit->{unitUsed}->{scaleFactor};
   $value .= ' ' . $intUnit->{unit}->{string}
      if defined($intUnit->{unit}->{string});
   $value .= ' ' . $intUnit->{unit}->{numeric}
      if defined($intUnit->{unit}->{numeric});

   return $value;
}


# CorrectLabel
#     Applies the quirks to the label part
#
sub CorrectLabel {
   my ($label, $quirks) = @_;

   return
      if !$label;

   $label =~ s/\s+/ /g;       # replace multiple whitespace by a single one
   $label =~ s/^\s+//;        # replace leading...
   $label =~ s/\s+$//;        # ... and trailing whitespace
   $label =~ tr/-_/  /        # convert underscores and dashes to space is requested
      if $quirks->{dashtospace};
   if ( $quirks->{correctcase} ) {
      $label = lc($label);    # convert to all lowercase
      $label = "\u$label";    # start with initial uppercase
      $label =~ s#[/_-\s](.)# \u\1#g;    # convert every first character of a word to upcase
   } else {
      $label = "\u$label";     # Start with uppercase letter, always
   }

   return $label;
}

# CorrectValue
#     Applies the quirks to the value part
#
sub CorrectValue {
   my ($value, $quirks) = @_;

   $value =~ s/^\s+\n//s;     # remove leading ...
   $value =~ s/\n\s+$//s;     # ... and trailing whitespace

   return undef
      if !$value;

   $value = '<FRAGMENT>' . $value . '</FRAGMENT>'
      if $quirks->{htmlcontent};
         #or &Display::HTML::isHTML($value);

   return $value;
}


sub parseIndenting {
   my ($line) = @_;

   $line =~ s/\t/        /g;  # convert tabs to spaces
   if ( wantarray ) {
      $line =~ s/\s+$//;      # remove trailing white space
      $line =~ s/^(\s*)//;    # remove leading indent
      return (length($1), $line);
   } else {
      $line =~ m/^(\s*)/;     # count the number of spaces
      return length($1);
   }
}

# parseLabelValueTree
#     Parses a tree structure with key value pairs
#     Simple example: IAFA record
#     Full example: FGDC SUTRS record
#
sub parseLabelValueTree {
   my ($quirks, $lines, $s) = @_;

   my @ar = ();    # the elements to be returned
   my $cn = 0;     # number of lines processed
   my $nocolon_allowed = 1;

   my $curindent = &parseIndenting($lines->[$s]);

   for (my $i = $s; $i <= $#{ $lines }; $i++) {
      my ($indent, $line) = &parseIndenting($lines->[$i]);

      last           # shorter indent but not an empty line? End of block
         if $indent < $curindent and $line ne '';

      $cn++;

      next           # skip empty lines
         if $line eq '';

      if ( $indent > $curindent ) {          # longer indent -> subtree
         my ($subvalue, $sn) = &parseLabelValueTree($quirks, $lines, $i);

         $ar[0] = {label=>'(no label)'}      # create an array entry if none exists
            if $#ar == -1 and $subvalue;

         if ( ref($subvalue) ) {
            if ( $Z3950::debug and defined($ar[$#ar]->{ref}) ) {
               warn '== Overwriting a reference to a structure';
               if ( eval 'use Data::Dumper;' ) {
                  warn &Dumper($ar[$#ar]->{ref}) . '== with the following' . &Dumper($subvalue);
               }
            }
            $ar[$#ar]->{ref} = $subvalue;
         } elsif ( $subvalue ) {
            $subvalue = $ar[$#ar]->{value} . "\n" . $subvalue
               if $ar[$#ar]->{value};

            $ar[$#ar]->{value} = $subvalue;
         }

         $i += $sn-1;
         $cn +=  $sn-1;
         $nocolon_allowed = 1;

      } else {       # $indent == $curindent
         if ( $line =~ m#:# or $nocolon_allowed ) {
            my ($label, $value) = split m/\s*:\s*/, $line, 2;

            if ( $value and $label ) {
               push @ar, {label => &CorrectLabel($label, $quirks),
                          value => &CorrectValue($value, $quirks)};
            } elsif ( $label ) {
               push @ar, {label => &CorrectLabel($label, $quirks)};
            } elsif ( $value ) {
               push @ar, {value => &CorrectValue($value, $quirks)};
            }
            $nocolon_allowed = 0;            # next line must include a colon
         } else {

            # Woops, no colon will there should be one. That
            # indicates that we are in the middle of a block of text and
            # we should discard all the results up to now and just
            # reparse the whole thing as a block of text. We will
            # have to find the end of the block still though.

            $cn = 0;                         # reset count
            for (my $i = $s; $i <= $#{ $lines }; $i++, $cn++) {
               last
                  if &parseIndenting($lines->[$i]) < $indent and $lines->[$i] !~ m/^\s*$/;
            }
            my @lines2 = map { substr($lines->[$i], $indent) } $lines->[$i..($i+$cn-1)];
            @ar = ({value => &CorrectValue(join("\n", @lines2), $quirks)});
         }
      }
   }

   # Optimise the returned value for various cases. Strip off as much as possible
   #
   if ( $#ar == 0 ) {
      if ( !$ar[0]->{label} ) {
         if ( !$ar[0]->{value} and !$ar[0]->{ref} ) {
            warn 'No label, no value, no ref, no nothing?';
            return (undef, $cn);
         } elsif ( !$ar[0]->{value} and $ar[0]->{ref} ) {
            return ($ar[0]->{ref}, $cn);
         } elsif ( $ar[0]->{value} and !$ar[0]->{ref} ) {
            return ($ar[0]->{value}, $cn);
         } else {
            return ($ar[0], $cn);               # ref and value available
         }
      } else {
         if ( !$ar[0]->{value} and !$ar[0]->{ref} ) {
            return ($ar[0]->{label}, $cn);      # return label as a plain scalar
         } else {
            return ($ar[0], $cn);               # return a hash
         }
      }
   } else {
      return (($#ar == -1? undef:[@ar]), $cn);  # anonymise the hash
   }
}

# parseColonCentered
#     Parses SUTRS with each line centered around a colon:
#
#         TITLE: Max Havelaar, of De Koffiveilingen der
#                Nederlandsche Handelmaatschappij
#          ISBN: 9023409809
#        AUTHOR: Multatuli
#
#     Returns a hash with the (label, value) pairs
#
sub parseColonCentered {
   my ($quirks, @lines) = @_;

   my ($line, $currelem, @ar) = ();

   my $center = index $lines[0], ':';

   foreach $line ( @lines ) {
      my ($label, $value) = (substr($line, 0, $center), substr($line, $center+1));

      $label =~ s/^\s+//;
      $label =~ s/\s+$//;
      $value =~ s/^\s+//;
      $value =~ s/\s+$//;

      next
         unless $label or $value;


      if ( !$label ) {
         warn 'No previous element to attach to? ?!'
            if $Z3950::debug and !$currelem;
         $currelem->{value} .= "\n" . $value;
      } else {
         push @ar, $currelem
            if $currelem;

         $currelem = undef;
         $currelem->{label} = $label;
         $currelem->{value} = $value;
      }
   }

   push @ar, $currelem
      if $currelem;

   # Apply the quirks for this server to all the label,value pairs
   #
   map { ($_->{label} = &CorrectLabel($_->{label}, $quirks),
          $_->{value} = &CorrectValue($_->{value}, $quirks))} @ar;

   if ( $#ar == -1 ) {
      return undef;     # no fields found, total emptiness
   } elsif ( $#ar == 0 ) {
      return $ar[0];    # return the only element present
   } else {
      return [@ar];     # anonymise hash
   }
}


# parseDIF
#     Parses a SUTR record that is formatted according to the
#     DIF specification (GCMD/GSFC/NASA/USA).
#
sub parseDIF {
   my ($quirks, @lines) = @_;

   my @rlines = @lines;
   my $line;
   my @ar = ();

   while ( $#lines > -1 ) {
      $line = shift @lines;
      $line =~ s/^\s+//;
      $line =~ s/\s+$//;

      next
         if !$line;
      last
         if $line =~ m/^End_Group/i;

      if ( $line !~ m/^(\S+)\s*:\s*(.*)/ ) {
         # woops, it seems to be plain text. Start from scratch and shift off
         # all the lines consumed by the plain text group
         my $line = shift @rlines;
         my @lines = ();
         $line =~ s/\t/        /g;
         $line =~ m/^(\s*)/;
         my $l = length($1);
         while ( $line !~ m/^\s*End_Group/i ) {
            push @lines, $line;
            $line =~ s/\t/        /g;
            $line =~ m/^(\s*)/;
            $l = length($1)
               if length($1) < $l;
            $line = shift @rlines;
         }
         @lines = map { substr($_, $l) } @lines
            if $l > 0;
         return (&CorrectValue(join("\n", @lines), $quirks), @rlines);
      } else {
         my ($label, $value) = ($1, $2);
         if ( lc($label) eq 'group' ) {
            $label = &CorrectLabel($value, $quirks);        # the group name is used as the label
            ($value,@lines) = &parseDIF($quirks, @lines);
            if ( ref($value) ) {
               push @ar, {label=>$label, 'ref'=>$value};
            } else {
               push @ar, {label=>$label, value=>$value};
            }
         } else {
            $value = &CorrectValue($value, $quirks);
            $label = &CorrectLabel($label, $quirks);
            push @ar, {label=>$label, value=>$value};
         }
      }
   }

   return (($#ar == 0? $ar[0]:[@ar]), @lines);
}


# parseRecordSUTRS
#     Parses the hash as a SUTRS record
#     Returns a hash with its contents. Elements of hash: title or abstract
#
sub parseRecordSUTRS {
   my ($content, $quirks, $name) = @_;

   $content =~ s/\n+$//;      # remove trailing newlines (multiple NLs)

   my $multiline = ($content =~ m/\n/s);

   if ( $quirks->{keyvaluetree} and $multiline ) {
      my @lines = split(m/\n/s, $content);
      my ($hash, $n) = &parseLabelValueTree($quirks, \@lines, 0);
      if ( $Z3950::debug and $n != $#lines+1 ) {
         warn "Lines left over after parseLabelValueTree ($n out of " . ($#lines+1) . " done):\n";
         map { print STDERR $_ . "\n" } @lines[($n-1)..$#lines];
      }
      return $hash;
   } elsif ( $quirks->{coloncentered} and $multiline ) {
      return &parseColonCentered($quirks, split(m/\n/s, $content));
   } elsif ( $quirks->{dif} and $multiline ) {
      $quirks->{correctcase} = 1;
      $quirks->{dashtospace} = 1;
      my ($hash, @lines) = &parseDIF($quirks, split(m/\n/s, $content));
      if ( $Z3950::debug and $#lines != -1 ) {
         warn 'Lines left over after parseDIF:';
         map { print STDERR $_ . "\n" } @lines;
      }
      return $hash;
   } else {
      return &CorrectValue(($quirks->{htmlcontent}? $content:'<PRE>'.$content.'</PRE>'), $quirks);
   }
}


# parseRecordGRS1
#     parses a structure as a GRS1 record.
#     Returns a hash. elements are label,value pairs. Value may contain
#     a reference to a hash of label,value pairs, the start of a tree of hashes.

sub parseRecordGRS1 {
   my ($taggedelement, $quirks, $name, $schema, $tagtype) = @_;
   my ($element) = ();

   $tagtype = 1                  # set some defaults
      if !defined($tagtype);
   $schema = 'NONE'
      if !defined($schema);

   my @rlist = ();
   foreach $element ( @{ $taggedelement } ) {      # SeqOf taggedElement
      my ($label, $value, $ref) = ('', '', 0);

      $element->{tagType} = $tagtype        # set the tagSet to the default if not specified
         if !defined($element->{tagType});

      # FIXME tagOccurrence, what is it?
      if ( $Z3950::debug ) {
         warn "-- Ignoring tagOccurrence $element->{tagOccurrence}"
            if exists $element->{tagOccurrence};
         warn "-- Ignoring metaData $element->{metaData}"
            if exists $element->{metaData};
         warn "-- Ignoring appliedVariant $element->{appliedVariant}"
            if exists $element->{appliedVariant};

         # In the GRS1 definition, diagnostic is EXTERNAL instead of something proper.
         # Don't know what to look for.
         warn "-- Ignoring diagnostic $element->{diagnostic}"
            if exists $element->{diagnostic};
      }

      # shorthand
      #
      my $content = $element->{content};
      my $tv;
      $tv = $element->{tagValue}->{string}
         if exists $element->{tagValue}->{string};
      $tv = $element->{tagValue}->{numeric}
         if exists $element->{tagValue}->{numeric};

      # Specific behaviour of tagSets
      # FIXME should we zap all the entries for tagset M? It looks like metadata
      # (implying that we get metadata on metadata, therefore doing metawork :-)

      if ( $element->{tagType} == 1 ) {
         next
            if $tv == 2 or $tv == 3 or $tv == 5 or $tv == 6;    # skip unnecessary tags

         if ( $tv == 1 ) {
            if ( !exists $content->{oid} ) {
               warn "Invalid Tagged Element in GRS1 record. tagType = but content is not OID"
            } else {
               if ( $content->{oid} eq $Z3950::OID_S_GILS ) {
                  $schema = 'GILS';
               } elsif ( $content->{oid} eq $Z3950::OID_S_CIP ) {
                  if ( !eval 'use Z3950::CIP;' ) {
                     warn 'Could not use() Z3950::CIP, ' . $!;
                  } else {
                     $schema = 'CIP';
                     warn '== Kinky stuff! CIP schema'
                        if $Z3950::debug;
                  }
               } else {
                  warn "== Kinky stuff! Unknown schema with OID '$content->{oid}'"
                     if $Z3950::debug;
               }
            }
         } elsif ( $tv == 4 ) {   #
            $tagtype = $content->{oid};
         }
         next;
      }

      next if exists $content->{elementNotThere}
           or exists $content->{elementEmpty}
           or exists $content->{noDataRequested};

      if ( exists $content->{subtree} ) {
         $ref = &parseRecordGRS1($content->{subtree}, $quirks, $name, $schema, $tagtype);
      } elsif ( exists $content->{intUnit} ) {
         $value = &ParseIntUnit($content->{intUnit});
      } else {
         $value = $content->{octets}
            if exists $content->{octets};
         $value = $content->{numeric}
            if exists $content->{numeric};
         $value = $content->{date}
            if exists $content->{date};
         $value = $content->{ext}
            if exists $content->{ext};
         $value = $content->{string}
            if exists $content->{string};
         $value = $content->{trueOrFalse}
            if exists $content->{trueOrFalse};
         $value = $content->{oid}
            if exists $content->{oid};
      }

      if ( $element->{tagType} == 1 ) {            # tagSet-M
         $label = $Z3950::tagSetM{$tv}
            or $label = "(M,$tv)";
      } elsif ( $element->{tagType} == 2 ) {       # tagSet-G
         $label = $Z3950::tagSetG{$tv}
            or $label = "(G,$tv)";
      } elsif ( $element->{tagType} == 4 and $schema eq 'GILS' ) {
         $label = $Z3950::tagSetGILS{$tv}
            or $label = "(GILS,$tv)";
      } elsif ( $element->{tagType} == 4 and $schema eq 'CIP' ) {
         $label = $Z3950::tagSetCIP{$tv}
            or $label = "(CIP,$tv)";
      } else {                                     # free-text or unknown tagset
         $label = "\u$tv";
         $element->{tagType} = 3
            if $Z3950::debug;
      }

      if ( 0 and $Z3950::debug ) {
         my @tagsets = ('', 'M', 'G', 'Local', $schema);
         my $l = $tv;
         $l = substr($l,0,5) . '...'
            if length($l) > 8;
         $label = "($schema,$tagsets[$element->{tagType}]-$l) $label";
      }

      my %hash = ();
      $hash{label} = &CorrectLabel($label, $quirks)
         if $label;
      $hash{value} = &CorrectValue($value, $quirks)
         if $value;
      $hash{ref} = $ref
         if $ref;
      push @rlist, { %hash };             # convert the hash into an anonymous ref and push the ref
   }

   return [ @rlist ];                     # convert the array into an anonymous ref to an array
}


sub parseRecordMarcLabel {
   my ($mapping, $mtype, $tag, $subfield) = @_;

   my $key = "$tag-$subfield";
   my $subkey = "$tag-$subfield-$mtype";

   if ( $mapping->{$subkey} ) {
      return $mapping->{$subkey};
   } elsif ( $mapping->{$key} ) {
      return $mapping->{$key};
   } else {
      return "($key)";
   }
}



# parseMarc($buffer [,$type])
#     Decodes Marc records in buffer into hash
#
#     $buffer  byte stream
#     $type    one of: 'USMARC', 'BLCMP' (default 'USMARC')
#
#     Returns ($fields, $pos) or $fields (depending on wantarray)
#     $fields contains a reference to an array, $pos points to the
#     first byte after the current record.
#
#  TODO Implement some kind of conversion from Marc fields to proper titles

sub parseRecordMarc {
   my ($buffer, $type, $quirks, $name) = @_;
   my $pos = 0;
   my %Leader;
   my $mapping = undef;
   my $mtype = undef;

   $type = 'USmarc'
      if !$type;
   $type = uc($type);

   my @s = unpack('A5 A A A A A A A A4 A A A A A A A A', substr($buffer,$pos,24));

   if ( $type eq 'USMARC' ) {    # specific
      %Leader = (
         lengthRecord         => $s[0],     # 0-4
         status               => $s[1],     # 5
         recordType           => $s[2],     # 6
         bibliographicLevel   => $s[3],     # 7
         controlType          => $s[4],     # 8
         indicatorCount       => $s[6],     # 10
         subfieldCodeCount    => $s[7],     # 11
         baseAddress          => $s[8],     # 12-16
         encodingLevel        => $s[10],    # 17
         cataloguingForm      => $s[11],    # 18
         linked               => $s[12],    # 19
         lengthLengthOf       => $s[13],    # 20
         lengthStartChar      => $s[14],    # 21
         lengthImplDef        => $s[15]     # 22
      );

      eval 'use Z3950::USMarc'
         or warn 'Could not use() Z3950::USMarc, ' . $!;

      $mapping = \%USMarc::mapping;

      if ( $Leader{recordType} =~ m/[acdefgijkmoprt]/ ) {
         $mtype = 'bibl';
      } elsif ( $Leader{recordType} =~ m/[z]/ ) {
         $mtype = 'auth';
      } elsif ( $Leader{recordType} =~ m/[vxy]/ ) {
         $mtype = 'hold';
      } elsif ( $Leader{recordType} =~ m/[w]/ ) {
         $mtype = 'clas';
      } elsif ( $Leader{recordType} =~ m/[q]/ ) {
         $mtype = 'comm';
      } else {
         $mtype = undef;
      }
   } elsif ( $type eq 'BLCMP' ) {
      %Leader = (
         lengthRecord         => $s[0],     # 0-4
         status               => $s[1],     # 5
         recordType           => $s[2],     # 6
         classRecord          => $s[3],     # 7
         indicatorCount       => $s[6],     # 10
         subfieldCodeCount    => $s[7],     # 11
         baseAddress          => $s[8],     # 12-16
         encodingLevel        => $s[10],    # 17
         analyticalIndicator  => $s[11],    # 18
         source               => $s[12],    # 19
         onUnion              => $s[13],    # 20
         SCPlength            => $s[14]     # 21
      );

      eval 'use Z3950::BLCMP'
         or warn 'Could not use() Z3950::BLCMP, ' . $!;

      $mapping = \%BLCMP::mapping;
   } else {
      die "Unknown version of MARC: '$type'";
   }

   my $pos += 24;       # Skip over leader (24 bytes)

   my $base = $pos;     # determine end of directory
   while (     $base < length($buffer)
           and ord(substr($buffer, $base, 1)) != 0x1e ) {
      $base += 12;
   }
   return ()            # should not happen, but then again, better safe than sorry
      if $base >= length($buffer);
   $base++;

   my (@fields, @rlist) = ();    # the list of fields/subfields found and the final return list

   # Read directory entries and the corresponding fields
   #
   while ( ord(substr($buffer, $pos, 1)) != 0x1e ) {
      my ($field, $level, $length, $start) = unpack('A3 A A3 A5', substr($buffer, $pos, 12));

      $length = $level . $length    # Only BLCMP has a per-field level
         if $type ne 'BLCMP';

      if ( $field =~ m/00./ ) {
         warn "-- Skipping Control Field number $field, '" . substr($buffer,$base+$start,$length-1) . "'"
            if $Z3950::debug;
      } else {
         my $indicator = substr($buffer, $base+$start, $Leader{indicatorCount});
         my @subfields = split m/\x1f/, substr($buffer, $base+$start+$Leader{indicatorCount},
                                           $length-$Leader{indicatorCount}-1);
         shift @subfields; # shift off the ghost subfield at the start of the field

         push @fields, map { { field=>$field, subfield=>substr($_,0,1), value=>substr($_,1) } } @subfields;
      }

      $pos += 12;
   }

   # Sort the fields according the (field, subfield) (primary, secondary key)
   @fields = sort { $a->{field} <=> $b->{field} or $a->{subfield} <=> $b->{subfield} } @fields;
   my $i;
   for ($i = 0; $i <= $#fields; ) {
      my ($field, $subfield) = ($fields[$i]->{field}, $fields[$i]->{subfield});

      # Convert the tag-char (field, subfield) into something more decent looking
      #
      my $label = &Z3950::parseRecordMarcLabel($mapping, $mtype, $field, $subfield);

      # Concatenate all the
      my @ar = ();
      while ( $i <= $#fields and $fields[$i]->{field} eq $field and $fields[$i]->{subfield} eq $subfield ) {
         push @ar, $fields[$i++]->{value};
      }

      push @rlist, { label=>$label, value=>($#ar==0? $ar[0]:undef), 'ref'=>($#ar==0? undef:[@ar]) };
   }

   return [ @rlist ];
}


# parseRecordSummary
#     Parses a structure as a Summary record
#     Returns a hash with label,value pairs.
# FIXME Untested record syntax parsing

sub parseRecordSummary {
   my ($content, $quirks, $name) = @_;

   warn 'Summary record encountered! Nice. It works doesn\'t it? No?! Why not ?!'
      if $Z3950::debug;

   return $content;
}


# parseRecordOPAC
#     Parses a structure as an OPAC record
#     Returns a hash with label,value pairs

sub parseRecordOPAC {
   my ($content, $quirks, $name) = @_;
   my ($element, $circ, $volume, $key, $key2);
   my (@rlist) = ();
   my (@ar, @ar2);

   warn 'bibliographicRecord not decoded'
      if $Z3950::debug and defined($content->{bibliographicRecord});

   my $n = 1;
   foreach $element ( @{ $content->{holdingsData} } ) {
      @ar = ();

      if ( $element->{marcHoldingsRecord} ) {
         my $value = &parseRecord($element->{marcHoldingsRecord}, $name.'(OPAC holding)');
         if ( ref($value) ) {
            push @ar, { label=>'Marc Info', 'ref'=>$value };
         } else {
            push @ar, { label=>'Marc Info', value=>$value };
         }
      }

      if ( $element->{holdingsAndCirc}->{circulationData} ) {
         @ar2 = ();
         foreach $circ ( @{ $element->{holdingsAndCirc}->{circulationData} } ) {
            foreach $key ( keys %{ $circ } ) {
               $key2 = "\u$key";
               $key2 =~ s/([a-z0-9])([A-Z])/\1 \2/g;
               push @ar2, { label=>$key2, value=>$circ->{$key} };
            }
         }
         push @ar, {label=>'Circulation Data', 'ref'=>[@ar2]}
            if $#ar2 != -1;

         @ar2 = ();
         foreach $volume ( @{ $element->{holdingsAndCirc}->{volumes} } ) {
            foreach $key ( keys %{ $volume } ) {
               $key2 = "\u$key";
               $key2 =~ s/([a-z0-9])([A-Z])/\1 \2/g;
               push @ar2, { label=>$key2, value=>$volume->{$key} };
            }
         }
         push @ar, {label=>'Volume', 'ref'=>[@ar2]}
            if $#ar2 != -1;
      }
      foreach $key ( keys %{ $element->{holdingsAndCirc} } ) {
         next if $key eq 'circulationData'
                 or $key eq 'volumes'
                 or $key eq 'marcHoldingsRecord';

         $key2 = "\u$key";
         $key2 =~ s/([a-z0-9])([A-Z])/\1 \2/g;
         push @ar, {label=>$key2, value=>$element->{$key}};
      }

      push @rlist, {label=>'Element '.$n++, 'ref'=>[@ar]};
   }

   return [ @rlist ];
}


=item parseRecord($record, $quirks [, $name])

Parses a structure (an \@EXTERNAL) of type indicated by its OID.

   $record     EXTERNAL
   $quirks     quirks used in decoding the structure, see InitConnection
   $name       symbolic name for warning messages

Returns a reference to a hash or a reference to an array with reference to hashes.
Each hash contains one or more of the following elements:

   label       label of the node in tree
   value       value attached to it
   ref         reference


=cut

sub parseRecord {
   my ($record, $quirks, $name) = @_;

   $name = 'record'
      if !$name;

   return undef
      unless exists $record->{encoding}
         and exists $record->{encoding}->{'single-ASN1-type'};

   my $content = $record->{encoding}->{'single-ASN1-type'};

   warn "-- $name to be parsed as ".&ASN1::OIDdescription($record->{'direct-reference'})
      if $Z3950::debug;

   if      ( $record->{'direct-reference'} eq $Z3950::OID_RS_SUTRS ) {
      return &parseRecordSUTRS($content, $quirks, $name);
   } elsif ( $record->{'direct-reference'} eq $Z3950::OID_RS_USmarc ) {
      return &parseRecordMarc($content, 'USmarc', $quirks, $name);
   } elsif ( $record->{'direct-reference'} eq $Z3950::OID_RS_GRS1 ) {
      return &parseRecordGRS1($content, $quirks, $name);
   } elsif ( $record->{'direct-reference'} eq $Z3950::OID_RS_Summary ) {
      return &parseRecordSummary($content, $quirks, $name);
   } elsif ( $record->{'direct-reference'} eq $Z3950::OID_RS_OPAC ) {
      return &parseRecordOPAC($content, $quirks, $name);
   } else {
      warn 'ParseRecord: Unsupported record format '.&ASN1::OIDdescription($record->{'direct-reference'}).' for '.$name;
      return undef
   }
}


=item diagnostic($diagnostic [, $defaultmsg])

Returns a human digestable string containing the diagnostic.
If no diagnostic is found or there was an error parsing it
the defaultmsg is returned.

   $diagnostic    DefaultDiagFormat or DiagRec

=cut

sub diagnostic {
   my ($diagnostic, $defaultmsg) = @_;

   # dereference DiagRec->defaultFormat
   #
   return &diagnostic($diagnostic->{defaultFormat})
      if exists $diagnostic->{defaultFormat};

   # fail on DiagRec->externallyDefined (is EXTERNAL)
   #
   return $defaultmsg . ' (externallyDefined diagnostic not decoded)'
      if exists $diagnostic->{externallyDefined};

   return $defaultmsg . ' (no diagnostic available)'
      if !$diagnostic;


   # Check out whether we have any useful additional info
   #
   my $addinfo;
   if ( exists $diagnostic->{addinfo}->{v2Addinfo} ) {
      $addinfo = $diagnostic->{addinfo}->{v2Addinfo};
   } else {
      $addinfo = $diagnostic->{addinfo}->{v3Addinfo};
   }


   # XXX we could put in a level of indirection to make the
   # set of diagnostics expandable (compare to recordsyntax)
   # instead of using %Bib1_Diagnostics directly.

   # the diagnostic set for Bib1 and Diag1 is the same
   my $diaginfo = $Z3950::Bib1_Diagnostics{$diagnostic->{condition}}
      if    $diagnostic->{diagnosticSetId} eq $Z3950::OID_D_Bib1
         or $diagnostic->{diagnosticSetId} eq $Z3950::OID_D_Diag1;


   # put a sensible string together
   #
   if ( $addinfo and $diaginfo ) {
      if ( $addinfo =~ m/$diaginfo/i ) {  # avoid similar messages
         return "$diagnostic->{condition}: $addinfo";
      } else {
         return "$diagnostic->{condition}: $diaginfo, $addinfo";
      }
   } elsif ( $diaginfo ) {
      return "$diagnostic->{condition}: $diaginfo";
   } elsif ( $addinfo ) {
      return "$diagnostic->{condition}: $addinfo";
   }

   return "$defaultmsg (error $diagnostic->{condition})";
}

=back

=cut

1;

__END__


