#!/usr/local/bin/perl

use Data::Dumper;

use Z3950;
use Z3950::Z3950;

# 'Use' the supported record syntaxes
#
use Z3950::ASN1::RecordSyntax_SUTRS;
use Z3950::ASN1::RecordSyntax_generic;

$::host = undef;              # remote host to query
$::port = 210;                # port the remote database is located on
$::dbase = undef;             # database to search in
$::string = undef;            # string to search on
$::inputtype = 'BER';         # default outputtype, can be XER or BER
$::fieldname = 'any';         # default field to search

$::debug = 0;                 # switch on/off debugging output and checks

while ( $arg = shift @ARGV ) {
   if ( $arg eq '-h' and $::host = shift @ARGV ) {
      # nop
   } elsif ( $arg eq '-p' and $::port = shift @ARGV ) {
      # nop
   } elsif ( $arg eq '-i' and $::inputtype = shift @ARGV ) {
      # nop
   } elsif ( $arg eq '-d' and $::dbase = shift @ARGV ) {
      # nop
   } elsif ( $arg eq '-s' and $::string = shift @ARGV ) {
      # nop
   } elsif ( $arg eq '-f' and $::fieldname = shift @ARGV ) {
      # nop
   } elsif ( $arg eq '-q' and $quirks = shift @ARGV ) {
      push @quirks, split /\s*,\s*/, $quirks;
   } elsif ( $arg eq '-D' and $debuglist = shift @ARGV ) {
      $ASN1::debug = $debug = 1
         if $debuglist =~ m/\bmain\b/i;
      $ASN1::debug_received = 1
         if $debuglist =~ m/\breceived\b/i;
      $ASN1::debug_sent = 1
         if $debuglist =~ m/\bsent\b/i;
      $ASN1::BER::debug = 1
         if $debuglist =~ m/\bBER\b/i;
      $ASN1::XER::debug = 1
         if $debuglist =~ m/\bXER\b/i;
      $Z3950::debug = 1
         if $debuglist =~ m/\bZ(39\.?50)?\b/i;
   } else {
      &usage();
   }
}

&usage("No host specified")
   unless $::host;
&usage("No port specified")
   unless $::port;
&usage("No database specified")
   unless $::dbase;
&usage("No search string specified")
   unless $::string;
&usage("No outputtype specified")
   unless $::inputtype;

foreach $quirk (@quirks) {
   $quirks->{dbases}->{$::dbase}->{$quirk} = 1;
}

# Initialise connection
#
$handle = &Z3950::InitConnection({'host'=>$::host, 'port'=>$::port, %{ $quirks }})
   or die $host.':'.$port.': '.$@.' (while initialising connection)';


# Prepare query
#
$field = &Z3950::Field($::fieldname);
$term  = &Z3950::RPNOperand($string, $field);
$query = &Z3950::RPNCreate($Z3950::OID_AS_Bib1, $term); 

# Do search and retrieve results if possible
#
print "Searching $::host:$::port:$::dbase for '$string' in $::fieldname field...\n";
$SR = &Z3950::Search($handle, $::dbase, $query, 1, 10)
   or die $::host.':'.$::port.': '.$@.' (while searching)';

if ( !$SR->{searchStatus} ) {         # something wrong?
   my $errormsg = &Z3950::diagnostic($SR->{records}->{nonSurrogateDiagnostic},
                                     'Search failed');
   die "Diagnostic: $errormsg";
}

# Check for results
#
print "Number of results: $SR->{resultCount}\n";
print "Records returned:  $SR->{numberOfRecordsReturned}\n";
@Records = @{ $SR->{records}->{responseRecords} }
   if $SR->{records}->{responseRecords};
print "Records found in SR: ".($#Records+1)."\n";

$nresults = $SR->{resultCount};


# Go get 'em! Fetch the remaining result records.
#
$n = $nresults-@Records;            # number records to be fetched at once
while ( @Records < $nresults ) {
   print 'Fetching '.($#Records+2)."+$n...\n";
   if ( $PR = &Z3950::Present($handle, 0, @Records+1, $n) ) {
      if ( $PR->{presentStatus} eq 'failure'
           or $PR->{numberOfRecordsReturned} == 0 ) {
         # try again, maybe the Response PDU was just too big
         if ( $n > 1 ) {
            warn "Failed to fetch $n records in one go";
            $n = int($n/2)
               or $n = 1;
         } else {
            warn "Failed to fetch 1 record in one go, giving up";
            last;
         }
      } else {
         push @Records, @{ $PR->{records}->{responseRecords} };
         $n = $nresults - @Records
            if $n > $nresults - @Records;
         last
            if $n == 0;
      }
   } else {
      warn $host.':'.$port.': '.$@.' (while fetching results)';
      last;
   }
}

# Fetch the first result in full (if there are results)
#
if ( $nresults ) {
   print "Fetching 1+1 (Full)...\n";
   if ( $PR = &Z3950::Present($handle, 1, 1, 1) ) {
      if ( $PR->{presentStatus} eq 'failure'
            || $PR->{numberOfRecordsReturned} == 0 ) {
         warn "Failed to fetch full record";
      } else {
         $Records[0] = ${ $PR->{records}->{responseRecords} }[0];
      }
   } else {
      warn $host.':'.$port.': '.$@.' (while fetching 1 full record)';
   }
}

print "Closing connection...\n";
&Z3950::CloseConnection($handle);


# Unwrap the records
#

$n = $#Records + 1;
$n = 10
   if $n > 10;

$Data::Dumper::Varname = "Record";
$Data::Dumper::Indent = 1;

print 'Converting first '.$n.' record'.($n == 1? '':'s')."...\n";
$dbasequirks = $quirks->{dbases}->{$dbase};
for ( $i = 0; $i < $n; $i++ ) {
   warn "Record[$i] diagnostic: ".&Z3950::diagnostic($Records[$i]->{record}->{surrogateDiagnostic})
      if $Records[$i]->{surrogateDiagnostic};
   push @parsedrecords, &Z3950::parseRecord($Records[$i]->{record}->{retrievalRecord},
                                            $dbasequirks, "Record[$i]")
      if exists $Records[$i]->{record}->{retrievalRecord};
   print Dumper($Records[$i])
      if ! exists $Records[$i]->{record}->{retrievalRecord};
}

print "Dumping " . ($#parsedrecords+1) . " records...\n";
$Data::Dumper::Varname = "Record";
$Data::Dumper::Indent = 1;
print &Dumper(@parsedrecords);

print "Done.\n";



sub usage {
      my (@msgs) = @_;

      print STDERR join("\n", @msgs) . "\n"
         if @msgs;
      die "Syntax: $0 [options]\n"
        . " with  -D    a comma separated list of (main,received,sent,BER,XER,Z) []\n"
        . "       -h    name of host to contact [$::host]\n"
        . "       -p    port [$::port]\n"
        . "       -i    type of connection (XER,BER) [$::inputtype]\n"
        . "       -d    database on destination [$::dbase]\n"
        . "       -f    fieldname to search [$::fieldname]\n"
        . "       -s    string to search for in $::fieldname field on destination ['$::string']\n"
        . "       -q    database quirks, comma separated list\n"
        . "             (keyvaluetree,dashtospace,correctcase)\n";
}
